import "package:utilary/utilary.dart";
import "glyph.dart";
import "color.dart";
import "font.dart";
import "dart:html";

class Terminal {
  final int rows;
  final int cols;

  int get width => cols;
  int get height => rows;

  Vector get size => Vector(cols, rows);

  final Array2D<Glyph> _glyphs;
  final Array2D<Glyph> _changedGlyphs;

  final Font _font;
  final CanvasElement _canvas;
  final CanvasRenderingContext2D _context;

  final int _scale = window.devicePixelRatio.toInt();

  Color foreColor = Color.white;
  Color backColor = Color.black;

  factory Terminal(int width, int height, Font font, [CanvasElement canvas]) {
    if (canvas == null) {
      canvas = CanvasElement();
      document.body.append(canvas);
    }

    return Terminal._(height, width, font, canvas);
  }

  Terminal._(this.rows, this.cols, this._font, this._canvas)
      : _context = _canvas.getContext("2d"),
        _glyphs = Array2D(rows, cols),
        _changedGlyphs = Array2D(rows, cols, Glyph.clear) {
    int canvasWidth = cols * _font.charWidth;
    int canvasHeight = rows * _font.lineHeight;

    _canvas.width = canvasWidth * _scale;
    _canvas.height = canvasHeight * _scale;
    _canvas.style.width = "${canvasWidth}px";
    _canvas.style.height = "${canvasHeight}px";

    _context.font = "${_font.size * _scale}px ${_font.family}, monospace";
  }

  void _setGlyph(int x, int y, Glyph glyph) {
    if (x < 0) return;
    if (x >= cols) return;
    if (y < 0) return;
    if (y >= rows) return;

    if (_glyphs.get(x, y) != glyph)
      _changedGlyphs.set(x, y, glyph);
    else
      _changedGlyphs.set(x, y, null);
  }

  void drawGlyph(int x, int y, Glyph glyph) => _setGlyph(x, y, glyph);

  void drawString(int x, int y, String str, [Color fore, Color back]) {
    Glyph temp;

    fore ??= foreColor;
    back ??= backColor;

    for (int i = 0; i < str.length; i++) {
      temp = Glyph.fromCharCode(str.codeUnitAt(i), fore, back);

      drawGlyph(x + i, y, temp);
    }
  }

  void fill(int x, int y, int width, int height, [Color color]) {
    color ??= backColor;
    Glyph temp = Glyph(" ", foreColor, color);

    for (int i = x; i < width; i++)
      for (int j = y; j < height; j++) drawGlyph(i, j, temp);
  }

  void clear() => fill(0, 0, this.width, this.height, backColor);

  void _renderGlyph(int x, int y, Glyph glyph) {
    //Fill the background
    _context.fillStyle =
        glyph.back.cssString; //Set context's fill color as glyph's backcolor
    _context.fillRect(
        x * _font.charWidth * _scale,
        y * _font.lineHeight * _scale,
        _font.charWidth * _scale,
        _font.lineHeight * _scale); //Draw the glyph's background

    //Don't draw blank characters
    if (glyph.charCode == 0 || glyph.charCode == " ".codeUnitAt(0)) return;

    _context.fillStyle =
        glyph.fore.cssString; //Set context's color to glyph's fore color
    _context.fillText(
        String.fromCharCode(glyph.charCode),
        (x * _font.charWidth + _font.x) * _scale,
        (y * _font.lineHeight + _font.y) * _scale);
  }

  void render() {
    for (int y = 0; y < rows; y++) {
      for (int x = 0; x < cols; x++) {
        Glyph glyph = _changedGlyphs.get(x, y);

        if (glyph == null) continue;

        _renderGlyph(x, y, glyph);

        _changedGlyphs.set(x, y, null);
        _glyphs.set(x, y, glyph);
      }
    }
  }
}
