import "dart:async";
import "dart:html" as html;
import "package:utilary/utilary.dart";

import "key_bindings.dart";
import "key_input.dart";

import "terminal.dart";

class UserInterface<T> {
  KeyBindings<T> keyPress = KeyBindings<T>();

  Terminal _terminal;
  List<Screen<T>> _screens = [];
  bool _dirty = true; //Whether or not the UI needs to be refreshed

  StreamSubscription<html.KeyboardEvent> _keyDownSubscription;

  UserInterface([this._terminal]);

  bool get handlingInput => _keyDownSubscription != null;

  void _keyDown(html.KeyboardEvent e) {
    KeyInput key =
        KeyInput(e.keyCode, shift: e.shiftKey, alt: e.altKey, ctrl: e.ctrlKey);

    Screen<T> screen = _screens.last;

    T input = keyPress.find(key);

    if (input != null) {
      e.preventDefault();
      if (screen.handleInput(input))
        return; //Let the screen handle the input if it can
    }

    //if the screen doesn't handle the input pass it the lower-level key
    screen.keyDown(key);
    e.preventDefault();
  }

  void setTerminal(Terminal terminal) {
    if (terminal == null) return;

    bool resized = _terminal == null ||
        _terminal.width != terminal.width ||
        _terminal.height != terminal.height;

    _terminal = terminal;
    dirty();

    if (resized)
      for (int i = 0; i < _screens.length; i++)
        _screens[i].resize(terminal.size);
  }

  void push(Screen<T> screen) {
    if (screen == null) return;

    screen._bind(this);
    _screens.add(screen);
  }

  void pop([Object result]) {
    Screen<T> popped = _screens.removeLast();
    popped._unbind();

    _screens.last.activate(popped, result);
  }

  ///Set whether this UI can handle input
  set handlingInput(bool value) {
    if (value == handlingInput) return;

    if (!value) {
      _keyDownSubscription.cancel();
      _keyDownSubscription = null;
    } else
      _keyDownSubscription = html.document.body.onKeyDown.listen(_keyDown);
  }

  ///Called every frame if this UI is running
  void _tick(num requestID) {
    refresh();

    if (_running) html.window.requestAnimationFrame(_tick);
  }

  bool _running = false;

  bool get running => _running;

  ///Start the animation loop. Leave [running] off if you want to
  ///manually refresh the UI
  set running(bool value) {
    if (value == _running) return;

    _running = value;

    if (_running) html.window.requestAnimationFrame(_tick);
  }

  void dirty() => _dirty = true;

  void refresh() {
    for (int i = 0; i < _screens.length; i++) _screens[i].update();

    if (_dirty) _render();
  }

  void _render() {
    if (_terminal == null) return;

    _terminal.clear();

    //find the index of the top most non transparent screen
    int i;
    for (i = _screens.length - 1; i >= 0; i--)
      if (!_screens[i].isTransparent) break;

    //Render the top most non transparent screen and every other
    //screen above it
    for (; i < _screens.length; i++)
      _screens[i]
          .render(_terminal); //Make the screen draw itself onto the terminal

    _dirty = false;

    _terminal.render();
  }
}

class Screen<T> {
  UserInterface _ui;

  UserInterface get ui => _ui;

  ///Does this screen allow screens below it to be rendered?
  bool get isTransparent => false;

  void resize(Vector size) {}

  ///If the UI has a bind for a specific key it lets
  ///the screen handle the [input] if it can. If the [input]
  ///passed can be handled returns true, otherwise false.
  ///In case the sceen can't handle the [input], lower-level
  ///[keyDown] will be called
  bool handleInput(T input) => false;

  bool keyDown(KeyInput key) => false;

  //Called every animation frame
  void update() {}

  ///Render this screen onto [terminal]
  void render(Terminal terminal) {}

  ///Called whenever this screen gets on top of the screens' stack
  ///[result] may have been given when [popped] had left the stack
  void activate(Screen<T> popped, [Object result]) {}

  void dirty() {
    if (_ui == null) return;

    _ui.dirty();
  }

  void _bind(UserInterface ui) {
    assert(ui != null);

    _ui = ui;
    resize(ui._terminal.size);
  }

  void _unbind() {
    _ui = null;
  }
}
